! ------------------- Class definition for AbstractReferencingObjectPolicy
doit
Object subclass: 'AbstractReferencingObjectPolicy'
  instVarNames: #( manager)
  classVars: #()
  classInstVars: #()
  poolDictionaries: #()
  inDictionary: Globals
  options: #()

%

doit
AbstractReferencingObjectPolicy comment: 
'AbstractReferencingObjectPolicy is the abstract superclass for policy classes 
that are used to do the referencing object cleanup (i.e., replacing the 
duplicate object with the canonical object). Subclasses should implement 
#''canonicalizeReferencesIn:''. See ReferencingObjectPolicy for the
default implementation.'
%
doit
AbstractReferencingObjectPolicy category: 'Canonicalization'
%
! ------------------- Class definition for ReferencingObjectPolicy
doit
AbstractReferencingObjectPolicy subclass: 'ReferencingObjectPolicy'
  instVarNames: #( referencingClassHistories)
  classVars: #()
  classInstVars: #()
  poolDictionaries: #()
  inDictionary: Globals
  options: #()

%

doit
ReferencingObjectPolicy comment: 
'ReferencingObjectPolicy is the default policy used to do the referencing object cleanup.
In order to support manager cleanup, referencing (or container) classes must be 
specified to this policy using #''addRefencingClass:''.'
%
doit
ReferencingObjectPolicy category: 'Canonicalization'
%
! ------------------- Class definition for CanonicalObjectManager
doit
Object subclass: 'CanonicalObjectManager'
  instVarNames: #( referencingObjectPolicy registries sessionCacheStatOffset)
  classVars: #()
  classInstVars: #( default)
  poolDictionaries: #()
  inDictionary: Globals
  options: #()

%

doit
CanonicalObjectManager comment: 
'The Object Canonicalization Framework provides an optional way to reduce
duplicate instances of equivalent objects in your application. To accomplish
this goal you register a class as being canonicalizable (#''createRegistryFor:'')
and then you can request a canonical object (#''canonicalObjectFor:''). 

To avoid commit conflicts, if a canonical object is not yet in the registry, the 
framework allows overlapping transactions to see different objects as the 
canonical one, but subsequent transactions will see one of the earlier ones.
To ensure that only one object is used, send #''ultimateCanonicalObjectFor:''
and be prepared to handle commit conflicts.

An alternate way to canonicalize objects is to provide a set of objects to the
manager and have it scan them for references to canonicalizable objects
(#''cleanupAll:''). This would typically be used with objects found using a
repository scan (Repository>>#''listReferencesToInstancesOfClasses:toDirectory:'').

Because some objects should not have references arbitrarily replaced (e.g., 
IdentitySets), the framework uses a pluggable policy to decide which
referencing objects can be modified. The default ReferencingObjectPolicy 
bases the decision on the referencing object class.

A special case of object cleanup is #''cleanupWriteSet'' that scans the write
set for the current transaction and replaces references as indicated by the
policy. Thus, a properly configured manager could do appropriate cleanup 
before a commit.'
%
doit
CanonicalObjectManager category: 'Canonicalization'
%
! ------------------- Class definition for CanonicalObjectPolicy
doit
Object subclass: 'CanonicalObjectPolicy'
  instVarNames: #()
  classVars: #()
  classInstVars: #()
  poolDictionaries: #()
  inDictionary: Globals
  options: #()

%

doit
CanonicalObjectPolicy comment: 
'CanonicalObjectPolicy is internal to the Canonical Object Framework.
It may be that in the future we will allow pluggable policies for some
canonicalization behavior.'
%
doit
CanonicalObjectPolicy category: 'Canonicalization'
%
! ------------------- Class definition for CanonicalObjectRegistry
doit
Object subclass: 'CanonicalObjectRegistry'
  instVarNames: #( canonicalObjectPolicy manager potentialCanonicalObjects
                    registry)
  classVars: #()
  classInstVars: #()
  poolDictionaries: #()
  inDictionary: Globals
  options: #()

%

doit
CanonicalObjectRegistry comment: 
'CanonicalObjectRegistry is internal to the Canonical Object Framework.
There may be one registry for each ClassHistory and it holds a collection 
(a KeyValueDictionary) of the ultimate canonical objects and a collection 
of potentialCanonicalObjects (an RcIdentityBag). If an equivalent object 
is not found in the registry, then the framework scans the potentials for
an equivalent object (which could have been added by any session). If
an equivalent object is found it is returned; otherwise the new object is
added to the potentials and returned.

At some time when conflicts are considered unlikely, one can send 
#''canonicalizePotentialCanonicalObjects'' to the CanonicalObjectManager 
and the potentials will be added to the registry (improving subsequent
lookup performance).'
%
doit
CanonicalObjectRegistry category: 'Canonicalization'
%
! ------------------- Class definition for CanonicalTestCase
doit
TestCase subclass: 'CanonicalTestCase'
  instVarNames: #( manager remoteSession)
  classVars: #( Root)
  classInstVars: #()
  poolDictionaries: #()
  inDictionary: Globals
  options: #()

%

doit
CanonicalTestCase comment: 
''
%
doit
CanonicalTestCase category: 'Canonicalization'
%

! ------------------- Remove existing behavior from AbstractReferencingObjectPolicy
doit
AbstractReferencingObjectPolicy removeAllMethods.
AbstractReferencingObjectPolicy class removeAllMethods.
%
! ------------------- Class methods for AbstractReferencingObjectPolicy
! ------------------- Instance methods for AbstractReferencingObjectPolicy
set compile_env: 0
category: 'Initialization'
method: AbstractReferencingObjectPolicy
setManager: aManager

	manager := aManager.
%
set compile_env: 0
category: 'Subclass Responsibilities'
method: AbstractReferencingObjectPolicy
canonicalizeReferencesIn: aReferencingObject

	self subclassResponsibility: #'canonicalizeReferencesIn:'.
%

! ------------------- Remove existing behavior from ReferencingObjectPolicy
doit
ReferencingObjectPolicy removeAllMethods.
ReferencingObjectPolicy class removeAllMethods.
%
! ------------------- Class methods for ReferencingObjectPolicy
set compile_env: 0
category: 'other'
classmethod: ReferencingObjectPolicy
new

	^super new
		initialize;
		yourself.
%
! ------------------- Instance methods for ReferencingObjectPolicy
set compile_env: 0
category: 'Accessors'
method: ReferencingObjectPolicy
referencingClasses

	^referencingClassHistories
		inject: IdentitySet new
		into: [:sum :each | sum addAll: each. sum].
%
set compile_env: 0
category: 'Cleanup'
method: ReferencingObjectPolicy
canonicalizeReferencesIn: aReferencingObject

	| counter |
	(referencingClassHistories includes: aReferencingObject class classHistory) ifFalse: [^self].
	manager sessionCacheStatAt: 5 incrementBy: 1.		"count of referencing objects scanned during cleanup"
	aReferencingObject isInvariant ifTrue: [
		manager sessionCacheStatAt: 8 incrementBy: 1.	"count of invariant referencing objects scanned during cleanup"
		^self
	].
	counter := 0.
	1 to: aReferencingObject _primitiveSize do: [:i |
		(self _canonicalizeReferenceIn: aReferencingObject atOffset: i) ifTrue: [counter := counter + 1].
	].
	counter == 0 ifTrue: [^self].
	manager
		sessionCacheStatAt: 6 incrementBy: 1;				"count of referencing objects changed during cleanup"
		sessionCacheStatAt: 7 incrementBy: counter;		"count of references changed during cleanup"
		yourself.
%
category: 'Cleanup'
method: ReferencingObjectPolicy
_canonicalize: existingObject inReferencingObject: aReferencingObject atOffset: anInteger
	"Answer true if we made a change, otherwise false"

	| canonical |
	(canonical := manager canonicalObjectFor: existingObject) == existingObject ifTrue: [^false].
	self
		_updateReferenceIn: aReferencingObject
		atOffset: anInteger
		toCanonical: canonical.
	^true
%
category: 'Cleanup'
method: ReferencingObjectPolicy
_canonicalizeReferenceIn: aReferencingObject atOffset: anInteger
	"offset is based on _privateSize, so includes named and numbered;
	answer true if object was replaced, otherwise answer false"

	| existing |
	(existing := aReferencingObject _primitiveAt: anInteger) isSpecial ifTrue: [^false].
	^self
		_canonicalize: existing
		inReferencingObject: aReferencingObject
		atOffset: anInteger.
%
category: 'Cleanup'
method: ReferencingObjectPolicy
_updateReferenceIn: referencingObject atOffset: anInteger toCanonical: canonicalObject
	"might not handle some classes"

	referencingObject
		_primitiveAt: anInteger
		put: canonicalObject.
%
set compile_env: 0
category: 'Initialization'
method: ReferencingObjectPolicy
initialize

	referencingClassHistories := IdentitySet new.
%
set compile_env: 0
category: 'Public'
method: ReferencingObjectPolicy
addReferencingClass: aClass

	referencingClassHistories add: aClass classHistory.
%
category: 'Public'
method: ReferencingObjectPolicy
addReferencingClasses: aCollection

	aCollection do: [:each |
		referencingClassHistories add: each classHistory.
	].
%

! ------------------- Remove existing behavior from CanonicalObjectManager
doit
CanonicalObjectManager removeAllMethods.
CanonicalObjectManager class removeAllMethods.
%
! ------------------- Class methods for CanonicalObjectManager
set compile_env: 0
category: 'Instance Creation'
classmethod: CanonicalObjectManager
clearDefault
"
	CanonicalObjectManager clearDefault.
"
	default := nil.
%
category: 'Instance Creation'
classmethod: CanonicalObjectManager
default

	default ifNil: [default := self new].
	^default.
%
category: 'Instance Creation'
classmethod: CanonicalObjectManager
new

	^super new
		initialize;
		yourself.
%
! ------------------- Instance methods for CanonicalObjectManager
set compile_env: 0
category: 'Accessors'
method: CanonicalObjectManager
 canonicalClasses

	^registries keys 
		inject: IdentitySet new
		into: [:sum :each | sum addAll: each. sum].
%
category: 'Accessors'
method: CanonicalObjectManager
referencingObjectPolicy

	^referencingObjectPolicy
%
category: 'Accessors'
method: CanonicalObjectManager
referencingObjectPolicy: aPolicy

	referencingObjectPolicy := aPolicy.
	aPolicy setManager: self.
%
category: 'Accessors'
method: CanonicalObjectManager
_registries

	^registries.
%
set compile_env: 0
category: 'Canonical'
method: CanonicalObjectManager
canonicalObjectFor: anObject
	"returns a possible canonical object"

	| domain |
	domain := self
		_registryFor: anObject
		ifAbsent: [^anObject].
	^domain canonicalObjectFor: anObject.
%
category: 'Canonical'
method: CanonicalObjectManager
hasCanonicalObjectFor: anObject

	| domain |
	domain := self
		_registryFor: anObject
		ifAbsent: [^false].
	^domain hasCanonicalObjectFor: anObject.
%
category: 'Canonical'
method: CanonicalObjectManager
hasUltimateCanonicalObjectFor: anObject

	| domain |
	domain := self
		_registryFor: anObject
		ifAbsent: [^false].
	^domain hasUltimateCanonicalObjectFor: anObject.
%
category: 'Canonical'
method: CanonicalObjectManager
isCanonicalObject: anObject
	"Could answer true for two equivalent objects if both are in potentialCanonicalObjects"

	| domain |
	domain := self
		_registryFor: anObject
		ifAbsent: [^false].
	^domain isCanonicalObject: anObject.
%
category: 'Canonical'
method: CanonicalObjectManager
isUltimateCanonicalObject: anObject

	| domain |
	domain := self
		_registryFor: anObject
		ifAbsent: [^false].
	^domain isUltimateCanonicalObject: anObject.
%
category: 'Canonical'
method: CanonicalObjectManager
ultimateCanonicalObjectFor: anObject
	"returns an actual canonical object (with possible conflicts)"

	| domain |
	domain := self
		_registryFor: anObject
		ifAbsent: [^anObject].
	^domain ultimateCanonicalObjectFor: anObject.
%
category: 'Canonical'
method: CanonicalObjectManager
_registryFor: anObject ifAbsent: aBlock

	^registries
		at: anObject class classHistory
		ifAbsent: aBlock.
%
set compile_env: 0
category: 'Cleanup'
method: CanonicalObjectManager
cleanupAll: aCollection

	| beginTime endTime |
	beginTime := System timeNs.
	aCollection do: [:each | 
		referencingObjectPolicy canonicalizeReferencesIn: each.
	].
	endTime := System timeNs.
	self 
		sessionCacheStatAt: 3 		incrementBy: 1;  						"count of calls to #'cleanupAll:'"
		sessionCacheStatAt: 4 		incrementBy: aCollection size;  	"count of objects scanned during cleanup"
		sessionCacheStatAt: 10	incrementBy: ((endTime - beginTime) // 1000 max: 1);	"microseconds in #'cleanupAll:'"
		yourself.
%
category: 'Cleanup'
method: CanonicalObjectManager
cleanupWriteSet

	| beginTime writeSet endTime |
	beginTime := System timeNs.
	writeSet := self _writeSet.
	endTime := System timeNs.
	self 
		sessionCacheStatAt: 9 incrementBy: ((endTime - beginTime) // 1000 max: 1);	"microseconds in building writeSet"
		cleanupAll: writeSet;
		yourself.
%
category: 'Cleanup'
method: CanonicalObjectManager
cleanupWriteSetAdding: newObjects ignoring: oldObjects

	| beginTime writeSet endTime |
	beginTime := System timeNs.
	writeSet := self _writeSetAdding: newObjects ignoring: oldObjects.
	endTime := System timeNs.
	self 
		sessionCacheStatAt: 9 incrementBy: ((endTime - beginTime) // 1000 max: 1);	"microseconds in building writeSet"
		cleanupAll: writeSet;
		yourself.
%
category: 'Cleanup'
method: CanonicalObjectManager
_writeSet

	| writeSet queue |
	queue := OrderedCollection withAll: System _writtenObjects.
	writeSet := IdentitySet new.
	[
		queue notEmpty.
	] whileTrue: [
		| next |
		next := queue removeLast.
		next class isBytes ifFalse: [
			writeSet add: next.
			1 to: next _primitiveSize do: [:i | 
				| x |
				x := next _primitiveAt: i.
				(x isSpecial or: [x class isBytes or: [x isCommitted or: [writeSet includes: x]]]) not ifTrue: [
					queue add: x.
				].
			].
		].
	].
	^writeSet.
%
category: 'Cleanup'
method: CanonicalObjectManager
_writeSetAdding: newObjects ignoring: oldObjects

	| writeSet queue |
	queue := OrderedCollection withAll: System _writtenObjects.
	writeSet := IdentitySet new.
	[
		queue notEmpty.
	] whileTrue: [
		| next |
		next := queue removeLast.
		next class isBytes ifFalse: [
			writeSet add: next.
			1 to: next _primitiveSize do: [:i | 
				| x |
				x := next _primitiveAt: i.
				(x isSpecial or: [x class isBytes or: [x isCommitted or: [writeSet includes: x]]]) not ifTrue: [
					queue add: x.
				].
			].
		].
	].
	^writeSet.
%
set compile_env: 0
category: 'Initialize'
method: CanonicalObjectManager
initialize

	registries :=  IdentityKeyValueDictionary new.
	self referencingObjectPolicy: ReferencingObjectPolicy new.
%
set compile_env: 0
category: 'Registry Management'
method: CanonicalObjectManager
canonicalizePotentialCanonicalObjects

	registries do: [:each | each canonicalizePotentialCanonicalObjects].
%
category: 'Registry Management'
method: CanonicalObjectManager
createRegistryFor: aClass

	registries
		at: aClass classHistory
		ifAbsentPut: [CanonicalObjectRegistry newForManager: self].
%
category: 'Registry Management'
method: CanonicalObjectManager
isCanonicalClass: aClass

	^registries includesKey: aClass classHistory.
%
set compile_env: 0
category: 'Statistics'
method: CanonicalObjectManager
resetSessionCacheStats

	| endOffset |
	sessionCacheStatOffset ifNil: [^self].
	endOffset := sessionCacheStatOffset + self sessionCacheStats size - 1.
	sessionCacheStatOffset to: endOffset do: [:i | 
		System sessionCacheStatAt: i put: 0.
	].
%
category: 'Statistics'
method: CanonicalObjectManager
sessionCacheStatAt: anIndex

	sessionCacheStatOffset ifNil: [^0].
	^System 
		sessionCacheStatAt: sessionCacheStatOffset + anIndex - 1.
%
category: 'Statistics'
method: CanonicalObjectManager
sessionCacheStatAt: anIndex incrementBy: aSmallInt

	sessionCacheStatOffset ifNil: [^self].
	System 
		sessionCacheStatAt: sessionCacheStatOffset + anIndex - 1
		incrementBy: aSmallInt.
%
category: 'Statistics'
method: CanonicalObjectManager
sessionCacheStatAt: anIndex put: aSmallInt

	sessionCacheStatOffset ifNil: [^self].
	System 
		sessionCacheStatAt: sessionCacheStatOffset + anIndex - 1
		put: aSmallInt.
%
category: 'Statistics'
method: CanonicalObjectManager
sessionCacheStatsSize

	^self sessionCacheStatDescriptions size
%
category: 'Statistics'
method: CanonicalObjectManager
sessionCacheStatDescriptions

	^#(
	"1"	'count of objects added to global registry'
	"2"	'count of objects added to potentialCanonicalObjects'

	"3"	'count of calls to #''cleanupAll:'''
	"4"    'count of objects scanned during cleanup'
	"5"    'count of referencing objects scanned during cleanup'
	"6"    'count of referencing objects changed during cleanup'
	"7"    'count of references changed during cleanup'
	"8"	'count of invariant referencing objects scanned during cleanup'

	"9"    'microseconds in building writeSet'
	"10"	'microseconds in #''cleanupAll:'''
	)
%
category: 'Statistics'
method: CanonicalObjectManager
sessionCacheStatOffset
	"nil means do not collect any statistics; aSmallInteger means start statistics here"

	^sessionCacheStatOffset
%
category: 'Statistics'
method: CanonicalObjectManager
sessionCacheStatOffset: anIntegerOrNil
	"nil means do not collect any statistics; aSmallInteger means start statistics here"

	sessionCacheStatOffset := anIntegerOrNil.
%
category: 'Statistics'
method: CanonicalObjectManager
sessionCacheStats
"
	CanonicalObjectManager soleInstance sessionCacheStats.
"
	^self sessionCacheStatsForSession: System session.
%
category: 'Statistics'
method: CanonicalObjectManager
sessionCacheStatsForSession: anInteger

	sessionCacheStatOffset ifNil: [^#()].
	^(System sessionCacheStatsForSessionId: anInteger)
		copyFrom: sessionCacheStatOffset + 1
		to: sessionCacheStatOffset + self sessionCacheStatDescriptions size.
%

! ------------------- Remove existing behavior from CanonicalObjectPolicy
doit
CanonicalObjectPolicy removeAllMethods.
CanonicalObjectPolicy class removeAllMethods.
%
! ------------------- Class methods for CanonicalObjectPolicy
! ------------------- Instance methods for CanonicalObjectPolicy
set compile_env: 0
category: 'other'
method: CanonicalObjectPolicy
canonicalRepresentationOf: anObject

	^anObject immediateInvariant.
%
category: 'other'
method: CanonicalObjectPolicy
wantsToCanonicalize: anObject

	^anObject size < 256.
%

! ------------------- Remove existing behavior from CanonicalObjectRegistry
doit
CanonicalObjectRegistry removeAllMethods.
CanonicalObjectRegistry class removeAllMethods.
%
! ------------------- Class methods for CanonicalObjectRegistry
set compile_env: 0
category: 'Instance Creation'
classmethod: CanonicalObjectRegistry
new

	self shouldNotImplement: #'new'.
%
category: 'Instance Creation'
classmethod: CanonicalObjectRegistry
newForManager: aManager

	^super new
		initializeForManager: aManager;
		yourself.
%
! ------------------- Instance methods for CanonicalObjectRegistry
set compile_env: 0
category: 'Accessors'
method: CanonicalObjectRegistry
_canonicalObjectPolicy

	^canonicalObjectPolicy.
%
category: 'Accessors'
method: CanonicalObjectRegistry
_potentialCanonicalObjects

	^potentialCanonicalObjects.
%
category: 'Accessors'
method: CanonicalObjectRegistry
_registry

	^registry.
%
set compile_env: 0
category: 'Initialize'
method: CanonicalObjectRegistry
initializeForManager: aManager

	canonicalObjectPolicy := CanonicalObjectPolicy new.
	manager := aManager.
	potentialCanonicalObjects := RcIdentityBag new.
	registry := KeyValueDictionary new.
%
set compile_env: 0
category: 'Lookup'
method: CanonicalObjectRegistry
canonicalObjectFor: anObject
	"returns an actual or potential canonical object (without 
	conflicts, but might need to canonicalize potentials later). 
	Does not yet deal with local registry."

	| canonical |
	canonical := self ultimateCanonicalObjectOrNilFor: anObject.
	canonical ifNotNil: [^canonical].
	potentialCanonicalObjects do: [:each | anObject = each ifTrue: [^each]].
	(canonicalObjectPolicy wantsToCanonicalize: anObject) ifFalse: [^anObject].
	canonical := canonicalObjectPolicy canonicalRepresentationOf: anObject.
	potentialCanonicalObjects add: canonical.
	manager sessionCacheStatAt: 2 incrementBy: 1.
	^canonical.
%
category: 'Lookup'
method: CanonicalObjectRegistry
hasCanonicalObjectFor: anObject

	| canonical |
	canonical := self ultimateCanonicalObjectOrNilFor: anObject.
	canonical ifNotNil: [^true].
	^potentialCanonicalObjects includesValue: anObject.
%
category: 'Lookup'
method: CanonicalObjectRegistry
hasUltimateCanonicalObjectFor: anObject

	"^(registry at: anObject otherwise: nil) notNil."
	^registry includesKey: anObject
%
category: 'Lookup'
method: CanonicalObjectRegistry
isCanonicalObject: anObject
	"Could answer true for two equivalent objects if both are in potentialCanonicalObjects"

	| canonical |
	canonical := self ultimateCanonicalObjectOrNilFor: anObject.
	canonical == anObject ifTrue: [^true].
	^potentialCanonicalObjects includesIdentical: anObject.
%
category: 'Lookup'
method: CanonicalObjectRegistry
isUltimateCanonicalObject: anObject

	^(registry at: anObject otherwise: nil) == anObject.
%
category: 'Lookup'
method: CanonicalObjectRegistry
ultimateCanonicalObjectFor: anObject
	"returns an actual canonical object (with possible conflicts)"

	^registry
		at: anObject
		ifAbsentPut: [
			manager sessionCacheStatAt: 1 incrementBy: 1.
			canonicalObjectPolicy canonicalRepresentationOf: anObject].
%
category: 'Lookup'
method: CanonicalObjectRegistry
ultimateCanonicalObjectOrNilFor: anObject

	^registry
		at: anObject
		ifAbsent: [nil].
%
set compile_env: 0
category: 'Registry Management'
method: CanonicalObjectRegistry
canonicalizePotentialCanonicalObjects

	| list |
	list := potentialCanonicalObjects asArray.
	potentialCanonicalObjects removeAll: list.
	list do: [:each | 
		self ultimateCanonicalObjectFor: each.
	].
%

! ------------------- Remove existing behavior from CanonicalTestCase
doit
CanonicalTestCase removeAllMethods.
CanonicalTestCase class removeAllMethods.
%
! ------------------- Class methods for CanonicalTestCase
set compile_env: 0
category: 'other'
classmethod: CanonicalTestCase
sampleDate

	^Date newDay: 1 year: 2014
%
! ------------------- Instance methods for CanonicalTestCase
set compile_env: 0
category: 'other'
method: CanonicalTestCase
setUp

	super setUp.
	CanonicalObjectManager clearDefault.
	manager := CanonicalObjectManager default.
	remoteSession := GsExternalSession newDefault.
	Root := Array new.
	System commit.
%
category: 'other'
method: CanonicalTestCase
tearDown

	Root := nil.
	remoteSession forceLogout.
	remoteSession := nil.
	manager := nil.
	CanonicalObjectManager clearDefault.
	super tearDown.
%
set compile_env: 0
category: 'tests'
method: CanonicalTestCase
test_canonicalClasses

	| list |
	self assert: manager canonicalClasses isEmpty.
	manager createRegistryFor: String.
	self 
		assert: (list := manager canonicalClasses) size == 2;
		assert: (list includes: String);
		assert: (list includes: MultiByteString);
		yourself.
%
category: 'tests'
method: CanonicalTestCase
test_canonicalObjectFor_01
	"Test local canonicalization"

	| date0 date1 date2 |
	date0 := CanonicalTestCase sampleDate.
	date1 := manager canonicalObjectFor: date0.
	self assert: date0 == date1.

	manager createRegistryFor: Date.
	date1 := manager canonicalObjectFor: date0.
	self assert: date0 == date1.
	date2 := manager canonicalObjectFor: date0 copy.
	self assert: date1 == date2.
%
category: 'tests'
method: CanonicalTestCase
test_canonicalObjectFor_02
	"Test conflict on simultaneous canonicalization"

	| oid date |
	manager createRegistryFor: Date.
	System commit.
	oid := remoteSession login; executeBlock: [
		| object |
		object := CanonicalObjectManager default canonicalObjectFor: CanonicalTestCase sampleDate.
		System commitTransaction.
		object asOop.
	].
	date := manager canonicalObjectFor: CanonicalTestCase sampleDate.
	self deny: date asOop == oid.
	System commit.
%
category: 'tests'
method: CanonicalTestCase
test_canonicalObjectFor_03
	"Test visibility of remote updates"

	| oid date |
	manager createRegistryFor: Date.
	System commit.
	oid := remoteSession login; executeBlock: [
		| object |
		object := CanonicalObjectManager default canonicalObjectFor: CanonicalTestCase sampleDate.
		System commitTransaction.
		object asOop.
	].
	System commit.
	date := manager canonicalObjectFor: CanonicalTestCase sampleDate.
	self assert: date asOop == oid.
%
category: 'tests'
method: CanonicalTestCase
test_canonicalObjectFor_04
	"Test with two objects in potentialCanonicalObjects"

	| oid date1 date2 |
	manager createRegistryFor: Date.
	System commit.
	oid := remoteSession login; executeBlock: [
		| object |
		object := CanonicalObjectManager default canonicalObjectFor: CanonicalTestCase sampleDate.
		System commitTransaction.
		object asOop.
	].
	date1 := manager canonicalObjectFor: CanonicalTestCase sampleDate.
	System commit.
	date2 := manager canonicalObjectFor: CanonicalTestCase sampleDate.
	self assert: (date1 == date2 or: [date2 asOop == oid]).
%
category: 'tests'
method: CanonicalTestCase
test_canonicalObjectFor_05
	"Test optimization benefit of global registry (avoiding the RcIdentityBag scan)"

	| date dates rcBag dict timeForRC timeForDict timeForRemote |
	manager createRegistryFor: Date.
	dates := Array new.
	rcBag := RcIdentityBag new.
	dict := KeyValueDictionary new.
	1 to: 1000 do: [:i | 
		date := (manager canonicalObjectFor: (Date newDay: i year: 2014)).
		dates add: date.
		rcBag add: date.
		dict at: date put: date.
	].
	manager canonicalizePotentialCanonicalObjects.
	System commit.
	timeForRC := System millisecondsToRun: [
		1 to: dates size do: [:i | 
			date := dates at: i.
			rcBag detect: [:each | each = date] ifNone: [].
		].
	].
	timeForDict := System millisecondsToRun: [
		1 to: dates size do: [:i | 
			date := dates at: i.
			dict at: date.
		].
	].
	timeForRemote := remoteSession login; executeBlock: [
		| list myManager |
		list := Array new.
		1 to: 1000 do: [:i | list add: (Date newDay: i year: 2014)].
		myManager := CanonicalObjectManager default.
		System millisecondsToRun: [
			1 to: 1000 do: [:i |
				myManager canonicalObjectFor: (list at: i).
			].
		].
	].
	self 
		assert: timeForDict < (timeForRC / 10);
		assert: (timeForRemote - timeForDict) abs < (timeForRemote - timeForRC) abs;
		yourself.
%
category: 'tests'
method: CanonicalTestCase
test_canonicalObjectFor_06
	"Avoid duplicate adds to global registry"

	| date1 oid date2 date3 |
	manager createRegistryFor: Date.
	System commit.
	date1 := CanonicalObjectManager default canonicalObjectFor: CanonicalTestCase sampleDate.
	oid := remoteSession login; executeBlock: [
		(CanonicalObjectManager default canonicalObjectFor: CanonicalTestCase sampleDate) asOop.
	].
	date2 := Object _objectForOop: oid.
	self assert: date2 isNil.
	manager canonicalizePotentialCanonicalObjects.
	System commit.
	date2 := Object _objectForOop: oid.
	self assert: date2 isNil.
	remoteSession commit.
	date2 := Object _objectForOop: oid.
	self assert: date2 isNil.
	System abort.
	date2 := Object _objectForOop: oid.
	self assert: date2 notNil.
	"now one in the global registry and one in potential"
	manager canonicalizePotentialCanonicalObjects.
	date3 := CanonicalObjectManager default canonicalObjectFor: CanonicalTestCase sampleDate.
	self assert: date1 == date3.
%
category: 'tests'
method: CanonicalTestCase
test_canonicalObjectFor_07
	"Ensure canonical object is invariant"

	| string1 string2 |
	manager createRegistryFor: String.
	string1 := 'DE' copy.
	self deny: string1 isInvariant.
	string2 := CanonicalObjectManager default canonicalObjectFor: string1.
	self 
		assert: string1 == string2;
		assert: string1 isInvariant;
		yourself.
%
category: 'tests'
method: CanonicalTestCase
test_canonicalObjectFor_08
	"Only canonicalize if size < 256"

	| string1 string2 string3 |
	manager createRegistryFor: String.
	string1 := (String new: 260) 
		atAllPut: $a;
		yourself.
	self deny: string1 isInvariant.
	string2 := CanonicalObjectManager default canonicalObjectFor: string1.
	string3 := CanonicalObjectManager default canonicalObjectFor: string1 copy.
	self 
		assert: string1 == string2;
		deny: string1 isInvariant;
		assert: string1 ~~ string3;
		yourself.
%
category: 'tests'
method: CanonicalTestCase
test_createRegistryFor_

	manager createRegistryFor: Date.
%
category: 'tests'
method: CanonicalTestCase
test_C_default

	| instance1 instance2 |
	instance1 := CanonicalObjectManager default.
	self assert: instance1 class == CanonicalObjectManager.
	instance2 := CanonicalObjectManager default.
	self assert: instance1 == instance2.
%
category: 'tests'
method: CanonicalTestCase
test_doCleanupOfWriteSet
	"Referenced object gets replaced"

	Root
		add: 1 -> CanonicalTestCase sampleDate;
		add: 2 -> CanonicalTestCase sampleDate;
		yourself.
	manager referencingObjectPolicy addReferencingClass: Association.
	manager createRegistryFor: Date.
	manager cleanupWriteSet.
	self assert: (Root at: 1) value == (Root at: 2) value.
%
category: 'tests'
method: CanonicalTestCase
test_doCleanupOn_01
	"Unregistered objects don't get replaced"

	Root
		add: 1 -> CanonicalTestCase sampleDate;
		add: 2 -> CanonicalTestCase sampleDate;
		yourself.
	manager referencingObjectPolicy addReferencingClass: Association.
	"manager createRegistryFor: Date."
	manager cleanupAll: Root copy.
	self assert: (Root at: 1) value ~~ (Root at: 2) value.
%
category: 'tests'
method: CanonicalTestCase
test_doCleanupOn_02
	"Referenced object gets replaced"

	Root
		add: 1 -> CanonicalTestCase sampleDate;
		add: 2 -> CanonicalTestCase sampleDate;
		yourself.
	manager referencingObjectPolicy addReferencingClasses: (Array with: Association).
	manager createRegistryFor: Date.
	manager cleanupAll: Root copy.
	self assert: (Root at: 1) value == (Root at: 2) value.
%
category: 'tests'
method: CanonicalTestCase
test_doCleanupOn_03
	"Canonicalize Date but not references from Association"

	Root
		add: 1 -> CanonicalTestCase sampleDate;
		add: 2 -> CanonicalTestCase sampleDate;
		yourself.
	"manager referencingObjectPolicy addReferencingClass: Association."
	manager createRegistryFor: Date.
	manager cleanupAll: Root copy.
	self assert: (Root at: 1) value ~~ (Root at: 2) value.
%
category: 'tests'
method: CanonicalTestCase
test_doCleanupOn_04
	"Referenced object gets replaced"

	Root
		add: 1 -> CanonicalTestCase sampleDate;
		add: (2 -> CanonicalTestCase sampleDate) immediateInvariant;
		yourself.
	manager referencingObjectPolicy addReferencingClasses: (Array with: Association).
	manager createRegistryFor: Date.
	manager cleanupAll: Root copy.
	self deny: (Root at: 1) value == (Root at: 2) value.
%
category: 'tests'
method: CanonicalTestCase
test_hasCanonicalObjectFor_

	| date |
	date := CanonicalTestCase sampleDate.
	manager createRegistryFor: Date.
	self
		assert: (manager canonicalObjectFor: date) == date;
		assert: (manager hasCanonicalObjectFor: date);
		assert: (manager hasCanonicalObjectFor: date copy);
		deny: (manager hasUltimateCanonicalObjectFor: date);
		yourself.
%
category: 'tests'
method: CanonicalTestCase
test_hasUltimateCanonicalObjectFor_

	| date |
	date := CanonicalTestCase sampleDate.
	manager createRegistryFor: Date.
	self
		assert: (manager ultimateCanonicalObjectFor: date) == date;
		assert: (manager hasCanonicalObjectFor: date);
		assert: (manager hasUltimateCanonicalObjectFor: date);
		assert: (manager hasUltimateCanonicalObjectFor: date copy);
		yourself.
%
category: 'tests'
method: CanonicalTestCase
test_isCanonicalClass_

	self deny: (manager isCanonicalClass: Date).
	manager createRegistryFor: Date.
	self assert: (manager isCanonicalClass: Date).
%
category: 'tests'
method: CanonicalTestCase
test_isCanonicalObject_

	| date |
	date := CanonicalTestCase sampleDate.
	manager createRegistryFor: Date.
	self
		assert: (manager canonicalObjectFor: date) == date;
		assert: (manager isCanonicalObject: date);
		deny: (manager isUltimateCanonicalObject: date);
		deny: (manager isCanonicalObject: date copy);
		yourself.
%
category: 'tests'
method: CanonicalTestCase
test_isUltimateCanonicalObject_

	| date |
	date := CanonicalTestCase sampleDate.
	manager createRegistryFor: Date.
	self
		assert: (manager ultimateCanonicalObjectFor: date) == date;
		assert: (manager isCanonicalObject: date);
		assert: (manager isUltimateCanonicalObject: date);
		deny: (manager isUltimateCanonicalObject: date copy);
		yourself.
%
category: 'tests'
method: CanonicalTestCase
test_referencingClasses

	| list |
	self assert: manager referencingObjectPolicy referencingClasses isEmpty.
	manager referencingObjectPolicy 
		addReferencingClass: Array;
		addReferencingClass: Association;
		yourself.
	self 
		assert: (list := manager referencingObjectPolicy referencingClasses) size == 2;
		assert: (list includes: Array);
		assert: (list includes: Association);
		yourself.
%
category: 'tests'
method: CanonicalTestCase
test_sessionCacheStats01
	"count of ultimate and potential"

	| stats |
	self assert: (stats := manager sessionCacheStats) = #().
	manager 
		createRegistryFor: Date;
		sessionCacheStatOffset: 0;
		resetSessionCacheStats;
		yourself.
	self 
		assert: (stats := manager sessionCacheStats) asSet asArray = #(0);
		assert: (manager canonicalObjectFor: CanonicalTestCase sampleDate) notNil;
		assert: 2 <= (stats := manager sessionCacheStats) size;
		assert: (stats at: 1) == 0;
		assert: (stats at: 2) == 1;
		yourself.
	manager canonicalizePotentialCanonicalObjects.
	self 
		assert: 2 <= (stats := manager sessionCacheStats) size;
		assert: (stats at: 1) == 1;
		assert: (stats at: 2) == 1;
		yourself.
%
category: 'tests'
method: CanonicalTestCase
test_sessionCacheStats02
	"cleanup statistics"

	| date stats |
	manager referencingObjectPolicy addReferencingClass: Association.
	manager 
		createRegistryFor: Date;
		sessionCacheStatOffset: 0;
		resetSessionCacheStats;
		yourself.
	self assert: (stats := manager sessionCacheStats) asSet asArray = #(0).
	date := manager ultimateCanonicalObjectFor: CanonicalTestCase sampleDate.
	self 
		assert: (manager isUltimateCanonicalObject: date);
		assert: (manager hasUltimateCanonicalObjectFor: CanonicalTestCase sampleDate);
		yourself.
	Root
		add: CanonicalTestCase sampleDate -> CanonicalTestCase sampleDate;
		add: (2 -> CanonicalTestCase sampleDate) immediateInvariant;
		add: (3 -> 33);
		yourself.
	manager cleanupWriteSet.
	self 
		assert: 10 == (stats := manager sessionCacheStats) size;
		assert: 1 == (stats at: 3);	"count of calls to cleanupAll:"
		assert: 0 <   (stats at: 4);	"count of objects scanned during cleanup"
		assert: 3 == (stats at: 5);	"count of referencing objects scanned during cleanup"
		assert: 1 == (stats at: 6);	"count of referencing objects changed during cleanup"
		assert: 2 == (stats at: 7);	"count of references changed during cleanup"
		assert: 1 == (stats at: 8);	"count of invariant referencing objects scanned during cleanup"
		assert: 0 <   (stats at: 9);	"microseconds in building writeSet"
		assert: 0 <  (stats at: 10);	"microseconds in #'cleanupAll:'"
		yourself.
%
category: 'tests'
method: CanonicalTestCase
test_ultimateCanonicalObjectFor_01
	"Test global canonicalization"

	| date0 date1 date2 |
	date0 := CanonicalTestCase sampleDate.
	manager createRegistryFor: Date.
	date1 := manager ultimateCanonicalObjectFor: date0.
	self assert: date0 == date1.
	date2 := manager ultimateCanonicalObjectFor: date0 copy.
	self assert: date1 == date2.
%
category: 'tests'
method: CanonicalTestCase
test_ultimateCanonicalObjectFor_02
	"Test optimization benefit of global registry (avoiding the RcIdentityBag scan)"

	| date dates rcBag dict timeForRC timeForDict timeForRemote |
	manager createRegistryFor: Date.
	dates := Array new.
	rcBag := RcIdentityBag new.
	dict := KeyValueDictionary new.
	1 to: 1000 do: [:i | 
		date := (manager ultimateCanonicalObjectFor: (Date newDay: i year: 2014)).
		dates add: date.
		rcBag add: date.
		dict at: date put: date.
	].
	"manager canonicalizePotentialCanonicalObjects."
	System commit.
	timeForRC := System millisecondsToRun: [
		1 to: dates size do: [:i | 
			date := dates at: i.
			rcBag detect: [:each | each = date] ifNone: [].
		].
	].
	timeForDict := System millisecondsToRun: [
		1 to: dates size do: [:i | 
			date := dates at: i.
			dict at: date.
		].
	].
	timeForRemote := remoteSession login; executeBlock: [
		| list myManager |
		list := Array new.
		1 to: 1000 do: [:i | list add: (Date newDay: i year: 2014)].
		myManager := CanonicalObjectManager default.
		System millisecondsToRun: [
			1 to: 1000 do: [:i |
				myManager canonicalObjectFor: (list at: i).
			].
		].
	].
	self 
		assert: timeForDict < (timeForRC / 10);
		assert: (timeForRemote - timeForDict) abs < (timeForRemote - timeForRC) abs;
		yourself.
%
