!=========================================================================
! Copyright (C) GemTalk Systems 1986-2016.  All Rights Reserved.
!
! $Id: profmonitor.gs 38924 2016-03-02 05:31:09Z lalmarod $
!
! Superclass Hierarchy:
!   ProfMonitor, Object.
!
!=========================================================================

expectvalue %String
run
Object _newKernelSubclass: 'ProfMonitor'
  instVarNames: #(#file #interval #results #sampleDepth #startTime #endTime 
		  #traceObjCreation #rawSampleArray )
  classVars: #( )
  classInstVars: #()
  poolDictionaries: { }
  inDictionary: Globals
    options: #() 
  reservedOop: 731
%

! remove existing behavior from ProfMonitor
removeallmethods ProfMonitor
removeallclassmethods ProfMonitor
set class ProfMonitor

category: 'For Documentation Installation only'
classmethod: 
installDocumentation

self comment: 
'ProfMonitor is a tool for taking snapshots of the execution stack
 at designated moments.  When done monitoring, the results are collected
 and formed into a report showing classes, method selectors, and hit rates.
 Execution stack snapshots can be gathered at either regular time intervals
 or programmatically by using System profMonSample.  Snapshots can also
 be taken during object creation by setting traceObjectCreation to true.
 
 The interval is the number of nanoseconds between sample points, which 
 must be either zero (for programmatic tracking) or >= 1000 nanoseconds. 
 Methods are provided to specify interval in milliseconds 
 (interval: variants) or in nanoseconds (intervalNs: variants) .

 Intervals less than 10 milliseconds use high resolution sampling
 and are implemented by reading a high resolution clock at each
 point where interrupts would be checked, (entries to non-trivial 
 non-primitive method, backward branches, and each CALL_PRIMITIVE bytecode).
 For the CALL_PRIMITIVE bytecodes , each CALL_PRIMITIVE in a loaded
 method is transformed to a PROFILING_CALL_PRIMITIVE (or to equivalent
 native code) at start of profiling and the transformation reversed 
 at end of profiling.  Methods loaded during profiling will have the
 transformation done at method load time.
 When using high resolution sampling, code executes at about 20%
 of normal execution speed.

 Intervals >= 10 milliseconds use the legacy timer interrupt implementation
 which has much less effect on execution performance.

 When the interval is zero, samples are not gathered based on clock time,
 but whenever System profMonSample is called.  Developers can use this
 to generate profiling information on specific methods of interest by
 adding calls to System profMonSample to these methods.
 
 The sampleDepth is the number of levels from the top of the stack to
 sample. This controls the depth of the stack for both method calls and
 object creation; it should be at least 20.
 
 The results are recorded in an instance of GsFile for the gem session,
 in the /tmp directory, in binary machine dependent form.  

 Instances of ProfMonitor should not be committed. Only instances
 newly created during life of a session are fully functional.

Instance Variables
------------------

file
  An instance of GsFile used to record sampling information while profiling
  is active.  The file contains binary data in machine dependent form.

interval
  The interval between sample points, in nanoseconds of CPU time.
  Must be either zero for programmatic sample gathering or >= 1000.

traceObjCreation 
  A Boolean. If true, object creation statistics are included in the profiling.

rawSampleArray 
  GemStone internal use, the total number of samples.

results
  Holds collected, processed snapshot information in instances of a
  class that is private to ProfMonitor.

sampleDepth 
  The number of levels from top of stack to sample.

startTime 
  Starting CPU time.

endTime
  Ending CPU time.
'.

%

! ------------------- Class methods for ProfMonitor

category: 'Defaults'
classmethod: 
defaultIntervalNs

"Returns the number of CPU nanoseconds used for a monitoring interval if no
 interval is given. "

 ^ 1000000 "1 millisecond"
%

category: 'Quick Profiling'
classmethod: 
monitorBlock: aBlock

"This is a quick way to profile the execution of a block and get a report of
 the result.  An interval of 1000 microseconds is used, and the results
 are reported down to 1 hit per method.  Returns a formatted report of the
 results of the profile run."

| inst result |
inst := self new .
inst monitorBlock: aBlock.
result := inst reportDownTo: 1 .
inst removeResults .
^ result
%

category: 'Quick Profiling'
classmethod: 
monitorBlock: aBlock downTo: hits 

"This is a quick way to profile the execution of a block and get a report of
 the result.  Returns a formatted report of the results of the profile run."

| inst result |
inst := self new .
inst monitorBlock: aBlock.
result := inst reportDownTo: hits .
inst removeResults .
^ result
%

category: 'Quick Profiling'
classmethod: 
monitorBlock: aBlock downTo: hits interval: millisecondsPerSample

"This is a quick way to profile the execution of a block and get a report of
 the result.  The msecsPerSample argument gives the CPU time interval between
 samples in milliseconds.  The hits argument gives the minimum number of hits
 for a method to be included in the report."

| inst result |
inst := self basicNew initializeNoFile .
inst _createFile: self newProfileFileName;  interval: millisecondsPerSample .
inst monitorBlock: aBlock.
result := inst reportDownTo: hits .
inst removeResults .
^ result
%

category: 'Quick Profiling'
classmethod: 
monitorBlock: aBlock downTo: hits intervalNs: nanosecondsPerSample

"This is a quick way to profile the execution of a block and get a report of
 the result.  The nanosecondsPerSample argument gives the CPU time interval between
 samples.  The hits argument gives the minimum number of hits
 for a method to be included in the report."

| inst result |
inst := self basicNew initializeNoFile .
inst _createFile: self newProfileFileName;  intervalNs: nanosecondsPerSample .
inst monitorBlock: aBlock.
result := inst reportDownTo: hits .
inst removeResults .
^ result
%

category: 'Quick Profiling'
classmethod: 
monitorBlock: aBlock intervalNs: nanosecondsPerSample

"This is a quick way to profile the execution of a block and get a report of
 the result.  The nanosecondsPerSample argument gives the CPU time interval 
 between samples.  The hits argument gives the minimum number of hits
 for a method to be included in the report."

| inst result |
inst := self basicNew initializeNoFile .
inst _createFile: self newProfileFileName;  intervalNs: nanosecondsPerSample .
inst monitorBlock: aBlock.
result := inst reportDownTo: 1 .
inst removeResults .
^ result
%

category: 'Profiling'
classmethod:
runBlock: aBlock intervalNs: nanosecondsPerSample

"Profile the execution of a block and return an instance
 of the receiver.  Caller must then send desired analysis selector
 to get the text results.
 The nanosecondsPerSample argument gives the CPU time interval between
 samples.  "

| inst |
inst := self basicNew initializeNoFile .
inst _createFile: self newProfileFileName;  intervalNs: nanosecondsPerSample .
[ inst startMonitoring.
  aBlock value.
] ensure:[ 
  inst stopMonitoring .
].
^ inst
%


category: 'Instance Creation'
classmethod: 
newWithFile: fileName

"Creates a new profiler with the given output file name and default monitoring
 interval."

^self newWithFile: fileName intervalNs: self defaultIntervalNs
%

category: 'Instance Creation'
classmethod: 
newWithFile: fileName interval: millisecondsPerSample

"Creates a new profiler with the specified output file name and monitoring
 interval"

| inst |
(inst := self basicNew) initializeNoFile ;
   _createFile: fileName; 
   interval: millisecondsPerSample .
^ inst
%
category: 'Instance Creation'
classmethod: 
newWithFile: fileName intervalNs: nanosecondsPerSample

"Creates a new profiler with the specified output file name and monitoring
 interval"

| inst |
(inst := self basicNew) initializeNoFile ;
   _createFile: fileName; 
   intervalNs: nanosecondsPerSample .
^ inst
%

category: 'Instance Creation'
classmethod: 
new

"Returns a new profiler with default initialization."

^ self basicNew initialize
%

category: 'Profiling'
classmethod: 
profileOn

"Creates a default instance, starts it monitoring, and returns it.
 To turn off profiling, send the message profileOff to the instance."

| inst |

inst := self new .
inst startMonitoring.
^inst.
%

category: 'Quick Profiling'
classmethod: 
spyOn: aBlock

"A convenience for Smalltalk programmers, this method merely fronts for
 ProfMonitor | monitorBlock:."

^self monitorBlock: aBlock
%

category: 'Private
classmethod: 
newProfileFileName

"Returns a String to be used as the name of a temporary file which will
 contain profiling data.  The file name will be /tmp/gem<processId>.pro,
 or gem<processId>.pro if a /tmp directory is not found."

| result pidStr aFile |
pidStr := (System gemVersionAt: #processId ) printString .
result := String withAll: '/tmp/gem'.
result  addAll: pidStr; addAll: '.pro' .

aFile := GsFile openWriteOnServer: result .
aFile ifNotNil:[
  aFile close .
  ^ result
  ].
result := (String withAll: 'gem').
result addAll: pidStr; addAll: '.pro' .
^ result
%

! _getMethodCounts deleted

! ------------------- Instance methods for ProfMonitor
category: 'Private'
method: 
forChild: childEntry tallyParentMethod: aSender rcvrClass: rcvrClass in: tempDict

"Private."

| children parentEntry d childTallies cTally parentTallies |
(aSender class == GsNMethod) ifFalse:[ 
  "ignore methods that were garbage collected and reused."
  ^ nil 
  ].

"Add parent to child"
(d := tempDict at: aSender otherwise: nil ) ifNil:[
	d := IdentityKeyValueDictionary new.
	tempDict at: aSender put: d ].

(parentEntry := d at: rcvrClass otherwise: nil) 
ifNil:[
  parentEntry := ProfMonitorEntry newForMethod: aSender 
                    receiverClass: rcvrClass.
  parentEntry total: 1.
  d at: rcvrClass put: parentEntry ] 
ifNotNil:[ 
  parentEntry incrementTotal ].

childTallies := parentEntry childTallies.
cTally := childTallies at: childEntry otherwise: 0.
childTallies at: childEntry put: (cTally + 1).

childEntry parents add: parentEntry.

"Add child to parent"
(children := parentEntry children ) ifNil:[
  children := IdentitySet new .
  parentEntry children: children .
].
children add: childEntry.

parentTallies := childEntry parentTallies.
cTally := parentTallies at: parentEntry otherwise: 0.
parentTallies at: parentEntry put: (cTally + 1).

^parentEntry
%

! fix 41385
category: 'Private'
method:
_intervalString: timeNs

| res divisor units |
res := 'monitoring interval: ' copy.
timeNs >= 1000 ifTrue:[
  divisor := 1000 . units := ' us' .
  timeNs >= 1000000 ifTrue:[ divisor := 1000000 . units := ' ms' ].
  res addAll: (timeNs // divisor ) asString ; add: $. ;
     addAll: (timeNs \\ divisor ) asString ; addAll: units .
] ifFalse:[
  res addAll: timeNs asString ; addAll: ' ns'
].
^ res
%

category: 'Private'
method: 
_stackReportDownTo: tallyThreshold

""

  | rpt lf nonrep nonreptallies total ave pct meths tallySet nameStr
    elapsedTime |

  tallySet := results at: 1 .
  nameStr := 'STATISTICAL STACK SAMPLING RESULTS' .
  rpt := String new .
  lf := Character lf.
  meths := tallySet sortDescending: 'total'.

  (endTime >= startTime) ifTrue: [
    elapsedTime := endTime - startTime.
  ] ifFalse: [
    elapsedTime := endTime + (16rffffffff - startTime).
  ].
  rpt add:'================'; add: lf ; add: nameStr; add: lf ;
    addAll: 'elapsed CPU time:    '; 
         addAll: (elapsedTime) asString; addAll: ' ms'; add: lf .
   
  rpt addAll: (self _intervalString: interval ) ; add: lf .

  rpt add: lf;
    addAll: ' total       %   class and method name'; add: lf;
    addAll: '------   -----   --------------------------------------'; add: lf.
  total := 0.
  meths do: [:each |
    each cmethod class == GsNMethod ifTrue:[
      total := total + each tally.
      ].
  ].

  nonrep := 0.  "non reported methods (below tally threshold."
  nonreptallies := 0.
  meths do: [:each |
    each cmethod class == GsNMethod ifTrue:[
      each total < tallyThreshold ifTrue: [
	      nonrep := nonrep + 1.
	      nonreptallies := nonreptallies + each tally.
      ] ifFalse: [ 
	      pct := each total asFloat * 100.0 / total.
	      rpt add: (each total asString width: -6); addAll: '  ' ;
				add: (pct asStringUsingFormat: #(-6 2 false)); addAll: '   ';
			  add: (each asStringWidth: 25);  add: lf .
      ].
    ].
  ].

  nonrep > 0 ifTrue: [
    pct := nonreptallies asFloat * 100.0 / total.
    rpt addAll: (nonreptallies asString width: -6); addAll: '  ' ;
        addAll: (pct asStringUsingFormat: #(-6 2 false)); addAll: '   ';
        addAll: nonrep asString ; addAll: ' other methods'; add: lf
  ].

  ave := total // (meths size max: 1).
  rpt addAll: (total asString width: -6); addAll: '  100.00   ';
    addAll: 'Total';
    add: lf.
  ^rpt
%

category: 'Updating'
method: 
interval: milliseconds

"Assign the sampling interval of the receiver."

| ns |
milliseconds _isSmallInteger ifFalse:[ ns _validateClass: SmallInteger ].
ns := milliseconds * 1000000 .
self intervalNs: ns .
%

classmethod
computeInterval: unprofiledCpuTimeSeconds
 "Returns an estimated profiling interval in nanoseconds for an execution
  taking the specified amount of cpu time, to yield approximately
  100000 samples . For 100000 samples, GEM_TEMPOBJ_CACHE_SIZE should
  be set to at least 300MB to avoid AlmostOutOfMemory errors during analysis 
  phase."
 | ns |
 ns := (5.0 * unprofiledCpuTimeSeconds / 100000 / 1.0e-9 ) asInteger .
 ns < 1000 ifTrue:[ ns := 1000 ].
 ^ ns
%

method: 
intervalNs: nanoseconds
  "Assign the sampling interval of the receiver."
nanoseconds _isSmallInteger ifFalse:[ nanoseconds _validateClass: SmallInteger ].
((nanoseconds < 1000) and: [nanoseconds ~~ 0]) ifTrue:[
  OutOfRange new name:'' min: 1000 actual: nanoseconds; 
   details: 'invalid sampling interval(nanoseconds)'; signal.
].
interval := nanoseconds
%

category: 'Accessing'
method: 
fileName

"Returns the name of the active file."

file ifNil:[ ^ nil ].
^ file name
%

category: 'Accessing'
method: 
interval

	^interval
%

category: 'Accessing'
method: 
startTime

	^startTime
%

category: 'Accessing'
method:
sampleDepth

        ^sampleDepth
%

category: 'Private'
method: 
_createFile: fileName

"Creates the sampling file for writing with the specified file name."
| f |
(f := file) ifNotNil:[ 
  f isOpen ifTrue:[ f close ].
  file := nil .
].
f := GsFile open: fileName mode: 'wb+' onClient: false .
f ifNil:[
  fileName _error: #bkupErrOpenFailed
              args:{ 'create' . GsFile serverErrorString }
].
file := f
%

category: 'Private'
method: 
_openFileAppend

"Reopens the sampling file for writing."

| f |
(f := file ) ifNil:[ self _error: 'cannot append to nil file'].
f isOpen ifFalse:[ | fileName |
  fileName := f pathName .
  f := GsFile open: fileName mode: 'ab+' onClient: false .
  f ifNil:[
    fileName _error: #bkupErrOpenFailed
              args:{ 'append' . GsFile serverErrorString }
  ].
  file := f
].
%

category: 'Private'
method: 
_openFileRead

"Reopens the sampling file for reading."

| f fileName |
(f := file ) ifNil:[ self _error: 'cannot reopen nil file'].
f isOpen ifTrue:[ 
  f seekFromBeginning: 0 .
] ifFalse:[
  fileName := f pathName .
  f := GsFile open: fileName mode: 'rb' onClient: false .
  f ifNil:[
    fileName _error: #bkupErrOpenFailed
                args:{ 'read' . GsFile serverErrorString }
  ].
  file := f 
].
%

category: 'Reporting'
method: ProfMonitor
gatherResults
"Analyze the receiver's file of sampling data and store the results of the
 analysis in the results instance variable of the receiver."

  | resultSet mthDict objDict rawArray j depth sysRepos rawArraySiz totalSamples elapsedTime |

  "Now process the statistical sample file."
  mthDict := IdentityKeyValueDictionary new .
  objDict := KeyValueDictionary new .
  rawArray := self _readSampleFile.
  rawArraySiz := rawArray size .
  sysRepos := SystemRepository .
  j := 1 .
  [j < rawArraySiz ] whileTrue:[  

    | sample recursionSet aMethod rcvrClass |
    sample := rawArray at: j .
    sample == sysRepos 
    ifFalse: [ self error: 'valid sample file: bad frame header' ]
    ifTrue:[   "sample header"
       j := j + 1 . 
       sample := rawArray at: j .
       sample _isSmallInteger 
       ifTrue:[ 
         depth := sample.  j := j + 1 ] 
       ifFalse:[
         self error:'invalid sample file: invalid depth' ]].

    recursionSet := IdentitySet new.
    aMethod := rawArray at: j .
    aMethod == true ifTrue:[ 

      "an object creation sample"
      | newObjClass aSender rcvr key d d2 anEntry child |
      newObjClass := rawArray at: j + 1 . 
      j := j + 2.
      aSender := rawArray at: j.
      rcvr := rawArray at: j + 1.
      rcvr := Object. "supress polymorphism reporting,"
                     "TODO add enable/disable "

      key := (Array with: newObjClass with: aSender).
      (d := objDict at: key otherwise: nil) ifNil: [
        d := KeyValueDictionary new.
        objDict at: key put: d. ].
      (d2 := d at: aSender otherwise: nil) ifNil: [
        d2 := IdentityKeyValueDictionary new.
        d at: aSender put: d2 ].

      (anEntry := d2 at: rcvr otherwise: nil) ifNil: [
        anEntry := ProfMonitorEntry newForClass: newObjClass
                        method: aSender receiverClass: rcvr.
        d2 at: rcvr put: anEntry ].
       anEntry incrementTally.
      "Now scan through call stack"
      child := anEntry.
      recursionSet add: child.
      1 to: depth-1 do: [:m |
        child ifNotNil: [ 
          aSender := rawArray at: j + (m * 2).
          rcvr := rawArray at: j + (m * 2) + 1.
          aSender ifNotNil: [ rcvr ifNotNil:[ 
            rcvr := Object .  "supress polymorphism reporting,"
                              "TODO add enable/disable "
            child := self forChild: child 
                          tallyParentMethod: aSender 
                          rcvrClass: rcvr 
                          in: d.
            (recursionSet includes: child) ifTrue: [
              child recursed: true].
            recursionSet add: child ]]]]]

    ifFalse:[ 

      "a normal statistical method execution sample"
      rcvrClass := rawArray at: j + 1.
      ((aMethod class == GsNMethod or:[ aMethod == #GCI ]) 
         and:[ rcvrClass isBehavior]) ifTrue:[ 
        | d anEntry |
        "method still exists or is a reenter marker"
        (d := mthDict at: aMethod otherwise: nil) ifNil: [
          d := IdentityKeyValueDictionary new.
          mthDict at: aMethod put: d. ].
        rcvrClass := Object .  "supress polymorphism reporting,"
                               "TODO add enable/disable "
        (anEntry := d at: rcvrClass otherwise: nil) ifNil:[
          anEntry := ProfMonitorEntry newForMethod: aMethod
                        receiverClass: rcvrClass.
          d at: rcvrClass put: anEntry ].
        anEntry incrementTally.
        depth > 1 ifTrue:[ | child rcvr |
          child := anEntry.
          recursionSet add: child.
          0 to: depth-2 do: [:m |
            child ifNotNil: [ | aSender |
              aSender := rawArray at: j + (m * 2) + 2.
              aSender ifNotNil: [
              rcvr := rawArray at: j + (m * 2) + 3.
              rcvr ifNotNil: [
                rcvr := Object .  "supress polymorphism reporting,"
                                  "TODO add enable/disable "
                child := self forChild: child 
                              tallyParentMethod: aSender 
                              rcvrClass: rcvr 
                              in: mthDict.
                (recursionSet includes: child) ifTrue: [
                  child recursed: true].
                recursionSet add: child. ]]]]]]].
   j := j + (depth * 2).
   ].

  (endTime >= startTime) ifTrue: [
    elapsedTime := endTime - startTime.
  ] ifFalse: [
    elapsedTime := endTime + (16rffffffff - startTime).
  ].

  "Consolidate results"
  results := Array new: 2 .
  resultSet := IdentityBag new.
  results at: 1 put: resultSet.
  totalSamples := 0.
  mthDict valuesDo: [:d | 
    d valuesDo: [ :anEntry |
      resultSet add: anEntry .
      anEntry cmethod class == GsNMethod ifTrue:[
        totalSamples := totalSamples + anEntry tally ]]].
  resultSet do: [:each | 
	  each totalSamples: totalSamples.
	  each elapsedTime: elapsedTime.
  ].
  results at: 2 put: objDict.
%

category: 'Private'
method: 
initialize

"Private."

self initializeNoFile .
self _createFile: self class newProfileFileName .
file close .
%

category: 'Private'
method: 
initializeNoFile

"Private."

interval := self class defaultIntervalNs .
sampleDepth := 1000.
traceObjCreation := false .
%

category: 'Monitoring'
method: 
monitorBlock: aBlock

"This method starts profiling, executes the block, and terminates profiling.
 Use reportDownTo: to generate results"

[ self startMonitoring.
  aBlock value.
] ensure:[
  self stopMonitoring .
].
self gatherResults 
%

category: 'Updating'
method: 
traceObjectCreation: aBoolean

"Enable (aBoolean == true) or disable profiling of object creation.  The state
 change will take effect on the next invocation of ProfMonitor>>startMonitoring
 for the receiver."

aBoolean _validateClass: Boolean .
traceObjCreation := aBoolean .
sampleDepth < 2 ifTrue:[ sampleDepth := 2 ].
%

category: 'Monitoring'
method: 
profileOff
 |rpt |
"Stop the given monitor and report."

self stopMonitoring; gatherResults; removeFile.
rpt := self reportDownTo: 1 .
^ rpt
%

category: 'Updating'
method: 
removeFile

"Removes the file generated by profiling operations in this profile monitor,
 if the file still exists."

| f fileName |
(f := file) ifNotNil:[
  fileName := file pathName .
  f close .
  GsFile removeServerFile: fileName .
  file := nil . 
].
%

category: 'Updating'
method: 
removeResults

"Releases results to aid garbage collection."

results ifNotNil:[
  results := nil .
  ].
%

category: 'Reporting'
method: 
report

"Formats and returns a string holding a report of the receiver's most
 recent profile run."

 ^ self reportDownTo: 1.
%

method:
reportAfterRun

"Formats and returns a string holding a report of the receiver's most
 recent profile run."
 | res |
 self gatherResults ; removeFile .
 res := self reportDownTo: 1.
 self removeResults .
 ^ res
%

method:
reportAfterRunDownTo: tallyThreshold

"Formats and returns a string holding a report of the receiver's most
 recent profile run."
 | res |
 self gatherResults ; removeFile .
 res := self reportDownTo: tallyThreshold.
 self removeResults .
 ^ res
%

category: 'Private'
method: 
_sendersReportDownTo: tallyThreshold

"Report formatting is:

  %     %                       Parent
 self total   total local  Method
 Time  Time    secs   %         Child

"

  | tallySet nameStr rpt lf meths elapsedTime totalSamples methodMs pMap cMap |
  tallySet := results at: 1 .
  nameStr := 'STATISTICAL METHOD SENDERS RESULTS' .
  rpt := String new .
  lf := Character lf.
  meths := tallySet sortDescending: 'total'.
  (endTime >= startTime) ifTrue: [
    elapsedTime := endTime - startTime.
  ] ifFalse: [
    elapsedTime := endTime + (16rffffffff - startTime).
  ].
  rpt add:'================'; add: lf ; add: nameStr; add: lf ;
    addAll: 'elapsed CPU time:    '; 
         addAll: (elapsedTime) asString; addAll: ' ms'; add: lf .
   
  rpt addAll: (self _intervalString: interval ) ; add: lf .

  totalSamples := 0.
  meths do: [:each | | aMeth |
    aMeth := each cmethod .
    (aMeth class == GsNMethod or:[ aMeth == #GCI ]) ifTrue:[
      totalSamples := totalSamples + each tally ]].

  rpt add: lf;
    addAll: '     %       %                     Parent'; add: lf;
    addAll: '  self  total   total  local  Method'; add: lf;
    addAll: '  Time   Time      ms    %         Child'; add: lf;
    addAll: '------ ------  ------  -----  -----------'; add: lf.
  meths do: [:each | | aMeth |
    aMeth := each cmethod .
    (aMeth class == GsNMethod or:[ aMeth == #GCI ]) ifTrue:[
      each total < tallyThreshold ifFalse:[ | parents children |
        rpt add: lf .
        methodMs := each total * elapsedTime asFloat / totalSamples.
        pMap := each parentTallies.
        cMap := each childTallies.
        "PARENTS"
        (parents := each parents) ifNotNil: [
          parents do: [:parent | | pt |
            pt := pMap at: parent otherwise: 0.
            rpt 
              add: '                 '; 
              add: ((pt * elapsedTime asFloat / totalSamples) 
                   asStringUsingFormat: #(-6 1 false)); space;
              add: ((pt * elapsedTime *100.0 / totalSamples / methodMs) 
                   asStringUsingFormat: #(-6 1 false)); space; space;
              add: '     ';
              add: (parent asStringWidth: 12) ; add: lf ]].
        "SELF"
        rpt add:'= ' ;
          add: ((each tally * 100.0 / totalSamples) 
               asStringUsingFormat: #(-6 1 false)); space;
          add: ((each total * 100.0  / totalSamples) 
               asStringUsingFormat: #(-6 1 false)); space; space;
          add: (methodMs asStringUsingFormat: #(-6 1 false)); space;
          add: ((each tally * elapsedTime *100.0 / totalSamples / methodMs) 
               asStringUsingFormat: #(-6 1 false)); space; space;
          add: (each asStringWidth: 12 ); add: lf . 
        "CHILDREN"
        (children := each children) notNil ifTrue: [
          children do: [:child | | ct |
            ct := cMap at: child otherwise: 0.
            rpt 
              add: '                 ';
              add: ((ct * elapsedTime asFloat / totalSamples) 
                   asStringUsingFormat: #(-6 1 false)); space;
              add: ((ct * elapsedTime * 100.0 / totalSamples / methodMs) 
                   asStringUsingFormat: #(-6 1 false)); space; space;
              add: '     ';
              add: (child asStringWidth: 12) ; add: lf ]].
        rpt addAll: '-----------------------------------------------------'; 
            add: lf ]]].
  ^rpt
%
!
! 40109: reimplement object creation profiling
!
!
category: 'Private'
method: ProfMonitor
_objCreationReportScanParents: entry level: level to: rpt seen: visited

  "Scan the parents of a given entry, printing them out indented 
   according to how deep in the scan we are.  Stop when we hit an
   entry we've already visited. "

  | parents tallies lf |

  lf := Character lf.
  (parents := entry parents) ifNotNil: [
    tallies := entry parentTallies.
    parents := (parents sortDescending: 'tally') asArray.
    parents do: [:parent | |  visit |
      visit := Array with: parent with: entry.
      (visited includes: visit) 
      ifFalse: [ 
        rpt add: '             '.
        level timesRepeat: [ rpt add: '  ' ].
        rpt 
          add: ((parent childTallies at: entry otherwise: -1) 
               asString width: -4);
          add: '  ';
          add: (parent asStringWidth: 12) ; add: lf.
          visited add: visit.
               self _objCreationReportScanParents: parent
                   level: (level + 1) to: rpt seen: visited ]]].
%

category: 'Private'
method: ProfMonitor
_objCreationReportDownTo: minTally

  | classes tallySet elapsedTime rpt lf nonrep nonreptallies meths counts |

  rpt := String new .
  lf := Character lf.

  (endTime >= startTime) 
  ifTrue: [elapsedTime := endTime - startTime] 
  ifFalse: [elapsedTime := endTime + (16rffffffff - startTime)].
  rpt add:'================'; lf ; 
      add: 'OBJECT CREATION REPORT:  '; lf ;
      add: 'elapsed CPU time:    '; 
      add: (elapsedTime) asString; add: ' ms';  lf ;
        add: (self _intervalString: interval ) ; lf ; lf.

  rpt add: lf;
    addAll: ' tally  class of created object'; add: lf.
  (sampleDepth > 1) 
  ifTrue: [
    rpt
    addAll: '           call stack'; add: lf.]
  ifFalse: [
    rpt
    addAll: '           callers'; add: lf.].

  nonrep := 0.  "non reported classes (below minTally threshold)"
  nonreptallies := 0.

  counts := IdentityKeyValueDictionary new.
  (results at: 2) keysAndValuesDo: [:k :d | | cls tally |
    cls := k at: 1.
    tally := (counts at: cls ifAbsent: [ counts at: cls put: 0 ]).
    d valuesDo: [:d2 |
      d2 valuesDo: [:e |
        e cclass isNil ifFalse: [ tally := tally + (e tally) ]]].
    counts at: cls put: tally ].
  classes := Array new.
  counts keysAndValuesDo: [:k :v | classes add: (Array with: k with: v)].
  classes := classes asSortedCollection: [:x :y | (x at: 2) > (y at: 2)].
  classes := classes collect: [:x | x at: 1].

  classes do: [:class |

    (counts at: class) < minTally ifTrue: [
      nonrep := nonrep + 1.
      nonreptallies := nonreptallies + (counts at: class).
      ]
    ifFalse:[  
      rpt
        addAll: '------  -----------------------------------------'; add: lf.

	  rpt add: lf ;
	    addAll: ((counts at: class) asString width: -6); addAll: '  ' ;
	    addAll: (class name asString width: 12) ; add: lf.

    tallySet := IdentitySet new.
    (results at: 2) keysAndValuesDo: [:k :d |
      ((k at: 1) = class) ifTrue: [ | sender |
        sender := (k at: 2).
        d valuesDo: [:d2 |
          d2 valuesDo: [:e | 
               e cclass isNil ifFalse: [
                 tallySet add: e]]]]].

    meths := tallySet sortDescending: 'tally'.
    meths do: [:each | | aMeth |
      aMeth := each cmethod .
      (aMeth _class == GsNMethod _or:[ aMeth == #GCI ]) ifTrue:[

        | aClass |
        aClass := aMeth .
        (sampleDepth > 1) ifTrue: [
          rpt addAll: ' - - -  - - - - - - - - - - - - - - - - - - - - -'; 
              add: lf ].
        rpt add: '         ';
            add: (each tally asString width: -6); 
            add: '  ';
            add: (each asStringWidth: 12) ; add: lf.
        self _objCreationReportScanParents: each level: 0 
             to: rpt seen: Set new ]]]].

  nonrep > 0 ifTrue: [
    rpt lf; lf;
        addAll: '------  -----------------------------------------'; add: lf;
        addAll: (nonreptallies asString width: -6); addAll: '  ' ;
        addAll: 'Instances of ' ;
        addAll: nonrep asString ; 
        addAll: ' other classes'; add: lf
  ].

  ^rpt
%

category: 'Private'
method: 
_invocationCountReportDownTo: tally

"Gemstone64 v2.0 , invocation counts not supported"

^ 'METHOD INVOCATION COUNTS - not implemented in this release' , Character lf
%

category: 'Private'
method: 
_samplingReportDownTo: tallyThreshold

""

  | rpt lf nonrep nonreptallies total ave pct meths tallySet nameStr
    elapsedTime |

  tallySet := results at: 1 .
  rpt := String new .
  nameStr := 'STATISTICAL SAMPLING RESULTS' .
  lf := Character lf.
  meths := tallySet sortDescending: 'tally'.

  (endTime >= startTime) ifTrue: [
    elapsedTime := endTime - startTime.
  ] ifFalse: [
    elapsedTime := endTime + (16rffffffff - startTime).
  ].
  rpt addAll:'================'; add: lf ; addAll: nameStr; add: lf ;
    addAll: 'elapsed CPU time:    '; 
         addAll: (elapsedTime) asString; addAll: ' ms'; add: lf .
   
  rpt addAll: (self _intervalString: interval ) ; add: lf .

  rpt add: lf;
    addAll: ' tally       %   class and method name'; add: lf;
    addAll: '------   -----   --------------------------------------'; add: lf.
  total := 0.
  meths do: [:each |
    each cmethod class == GsNMethod ifTrue:[
      total := total + each tally ]].

  nonrep := 0.  "non reported methods (below tally threshold."
  nonreptallies := 0.
  meths do: [:each |
    each cmethod class == GsNMethod ifTrue:[
      each tally < tallyThreshold ifTrue: [
	nonrep := nonrep + 1.
	nonreptallies := nonreptallies + each tally ]
      ifFalse: [ 
	pct := each tally asFloat * 100.0 / total.
	rpt add: (each tally asString width: -6); addAll: '  ' ;
	    add: (pct asStringUsingFormat: #(-6 2 false)); addAll: '   ';
	    add: (each asStringWidth: 25) ; add: lf ]]].

  nonrep > 0 ifTrue: [
    pct := nonreptallies asFloat * 100.0 / total.
    rpt addAll: (nonreptallies asString width: -6); addAll: '  ' ;
        addAll: (pct asStringUsingFormat: #(-6 2 false)); addAll: '   ';
        addAll: nonrep asString ; addAll: ' other methods'; add: lf ].

  ave := total // (meths size max: 1).
  rpt addAll: (total asString width: -6); addAll: '  100.00   ';
    addAll: 'Total';
    add: lf.
  ^rpt
%

category: 'Reporting'
method: 
reportDownTo: tallyThreshold

"Formats and returns a string holding a report of the receiver's most recent
 profile run.  Stops reporting when a tally smaller than tally is
 encountered."

  | rpt lf aReport |

  results ifNil: [^'No profiling data are available.'].
  
  lf := Character lf.
  rpt := String new .
  (self dynamicInstVarAt: #overrunStr) ifNotNil:[:str | rpt add: str ; lf].
  aReport := self _samplingReportDownTo: tallyThreshold  .
  rpt addAll: aReport;  add: lf ; add: lf .

  aReport := self _stackReportDownTo: tallyThreshold.
  rpt addAll: aReport;  add: lf ; add: lf .

  aReport := self _sendersReportDownTo: tallyThreshold . 
  rpt addAll: aReport;  add: lf ; add: lf .

  traceObjCreation ifTrue:[  
    aReport := self _objCreationReportDownTo: tallyThreshold .
    rpt addAll: aReport;  add: lf ; add: lf .
  ].

 "aReport := self _invocationCountReportDownTo: tallyThreshold .
  rpt addAll: aReport;  add: lf . "

  ^ rpt
%

category: 'Accessing'
method: 
endTime

	^endTime
%

category: 'Accessing'
method: 
results

"Returns the value of the instance variable results."

^results
%

category: 'Updating'
method: 
results: newValue

 "Modify the value of the instance variable results."

results:= newValue
%

category: 'Updating'
method:
sampleDepth: newValue

"Modify the value of the instance variable sampleDepth. Should be
 at least 20; small values will prune method call stacks and object
 creation stacks."

sampleDepth := newValue
%

category: 'Monitoring'
method: 
startMonitoring

"Starts monitoring."
self _openFileAppend .
self _zeroArgPrim: 1 .  "enable monitoring in the virtual machine"
startTime := System _readClock.
%

category: 'Monitoring'
method: 
stopMonitoring

"Stops monitoring. stores total number of samples written to file
  in the rawSampleArray instVar."
| numSamples delta |
endTime := System _readClock.
(delta := endTime - startTime) <= 0 ifTrue:[  "fix 41260"
  (OutOfRange new ) 
     name: 'CPU time used' min: 10 actual: delta ; 
     details: 'code to be profiled must use at least 10ms of CPU time' ; 
	 signal
].
numSamples := self _zeroArgPrim: 2  . "stop monitoring in the virtual machine"
rawSampleArray := numSamples .
self dynamicInstVarAt: #overrunStr put: (self _zeroArgPrim: 6).
file close .
%


category: 'Private'
method: 
_readSampleFile

"Returns the raw sample data generated by the monitoring code, for use
 by the output routines.

 Returns anArray in the following format:

           aSmallInteger - depth of sampling for next section of Array,
 optional: nil
 optional: aClass
           aGsNMethod 1
           aClass     1
           aGsNMethod 2
           aClass     2
           ...
           aGsNMethod N
           aClass     N

 where N = depth as specified by the initial aSmallInteger.

 The (aGsNMethod, aClass) pairs represent stack frames.
 The optional (nil, aClass) pair at the beginning indicate the sample 
 was taken during the creation of an instance of the designated class.  
"

| arr |
self _openFileRead.
arr := self _zeroArgPrim: 3.
file close .
^ arr
%

category: 'Private'
method: 
_zeroArgPrim: opcode

"opcode      function
   1		start profiling - returns receiver
   2		stop profiling - returns number of samples written to sample file
   3    read sample file - returns contents of file in an Array
   4    suspend high res sampling
   5    resume  high res sampling
   6    get and clear overrun info, result is a String or nil
"
<primitive: 191>
self _primitiveFailed: #_zeroArgPrim: args: { opcode }
%

category: 'Monitoring'
method:
suspendSampling
  "if high resolution sampling in use, suspends sampling."
interval >= 10000000 ifTrue:[
  self error:'suspend only supported with high resolution sampling'
].
self _zeroArgPrim: 4 
%

method:
resumeSampling
  "if high resolution sampling in use, resumes sampling."
interval >= 10000000 ifTrue:[
  self error:'resume only supported with high resolution sampling'
].
self _zeroArgPrim: 5 
%


